% CHEIRALITYCONSTRAINT2 Chooses the correct solution from 4 given ones.
%
% Usage: [R, C, i] = cheiralityConstraint(Rs, Cs, K1, K2, x1i, x2i)
%
%   Arguments:
%       Rs  - 1x4 cell array of rotation matrices
%       Cs  - 1x4 cell array of camera translation vectors
%       K1  - calibration matrix of first camera
%       K2  - calibration matrix of second camera
%       x1i - point correspondences in first image
%       x2i - point correspondences in second image
%
%   Returns:
%       R - chosen rotation matrix
%       C - chosen camera translation vector
%       i - index of correct solution
%
% Chooses the correct solution for R and C from 4 supplied solutions by
% testing the cheirality constraint.
%
% See Also: CHEIRALITYCONSTRAINT, POSESFROMESSENTIALMATRIX
%
% @author Isaac Esteban (FIT3D), Hanno Jaspers
% @date 2011
function [ Rcorr, Ccorr, correctSolution] = cheiralityConstraint2(R, C, K1, K2, x1i, x2i)

if nargin ~= 6 || ~iscell(R) || ~iscell(C)
    error 'Wrong (number of) arguments!'; 
end

PXcam = zeros(3,4,4);
for i=1:4
   PXcam(:,:,i) = [R{i}  -R{i}*C{i}];
end

%% Resolve ambiguous solution

    voting = zeros(4,1);
    votingB = zeros(4,1);

    % For every correspondence
    for i=1:size(x1i,2)

        % Two matching points in image coordinates (x in image 1 and xp in
        % image 2)
        x  = [x1i(1:2,i);1];
        xp = [x2i(1:2,i);1];

        % The first camera matrix is taken P = [I|0] and the other 
        Pcam = [eye(3),zeros(3,1)];
        P = K1*Pcam;
        xhat = K1\x;
        

        % For each camera matrix (PXcam), reproject the pair of points in 3D
        % and determine the depth in 3D of the point
        X3D = zeros(4,4);
        Depth = zeros(4,2);
        
        b = zeros(4,1);
        c = zeros(4,1);
        d = zeros(4,1);
        

        
        for p=1:4

            % First the point is converted
            xphat = K2\xp;

            % We build the matrix A
            A = [Pcam(3,:).*xhat(1,1)-Pcam(1,:);
                 Pcam(3,:).*xhat(2,1)-Pcam(2,:);
                 PXcam(3,:,p).*xphat(1,1)-PXcam(1,:,p);
                 PXcam(3,:,p).*xphat(2,1)-PXcam(2,:,p)];

            % Normalize A
            A1n = sqrt(sum(A(1,:).*A(1,:)));
            A2n = sqrt(sum(A(2,:).*A(2,:)));
            A3n = sqrt(sum(A(3,:).*A(3,:)));
            A4n = sqrt(sum(A(4,:).*A(4,:))); 
            Anorm = [A(1,:)/A1n;
                     A(2,:)/A2n;
                     A(3,:)/A3n;
                     A(4,:)/A4n];

            % Obtain the 3D point
            [Uan,San,Van] = svd(Anorm);
            X3D(:,i) = Van(:,end);

            % Check depth on second camera
            xi = PXcam(:,:,p)*X3D(:,i);
            w = xi(3);
            T = X3D(end,i);
            m3n = sqrt(sum(PXcam(3,1:3,p).*PXcam(3,1:3,p)));
            Depth(p,1) = (sign(det(PXcam(:,1:3,p)))*w)/(T*m3n);

            % Check depth on first camera
            xi = Pcam(:,:)*X3D(:,i);
            w = xi(3);
            T = X3D(end,i);
            m3n = sqrt(sum(Pcam(3,1:3).*Pcam(3,1:3)));
            Depth(p,2) = (sign(det(Pcam(:,1:3)))*w)/(T*m3n);

        end;


        % Voting
        %bT = 10e10;
        %dT = 10e10;
        %cT = 10e10;
        %voteC = 1;
        for p=1:size(PXcam,3)
            
            % Vote based on depth
            if(Depth(p,1)>0 && Depth(p,2)>0)
                voting(p) = voting(p)+1;
            end;
            
            
            % Vote for constraints (NOT USED)
            %if (b(p)+d(p)+c(p) <= cT+bT+dT)
            %if (b(p)*d(p)*c(p) < cT*bT*dT)
            %if (b(p) < bT && d(p) < dT && c(p) < cT)
            %if d(p) <= dT
            %    bT = b(p);
            %    dT = d(p);
            %    cT = c(p);
            %    voteC = p;
            %end;    
        end;


    
    end;
    voting
[maxv, maxi] = max(voting);
correctSolution = maxi(1);
    
Rcorr = R{correctSolution};
Ccorr = C{correctSolution};


end

