%DECOMPOSEHOMOGRAPHY Decomposes a homography into R, t and n
%   
% Usage: [ decomp ] = decomposeHomography( H )
%
% Arguments:
%       H - homography matrix
%
% Returns:
%       decomp - a struct (most likely 8x1) with the fields:
%           R - 3x3 rotation matrix of the plane
%           t - 3x1 translation vector of plane
%           n - 4x1 normal of plane
%           d, Rs, ts
%
% Decomposes homography using Faugeras SVD-based method.
%
% See also: chooseCorrectHomographyDecomposition
%
% @author Hanno Jaspers
% @date 2011
function [ decomp ] = decomposeHomography( H )

% perform SVD of H
[U S V] = svd(H);

% think of S as another homography matrix that can be written as 
% S = Rs + ts*ns' ...?

d1 = abs(S(1,1));
d2 = abs(S(2,2));
d3 = abs(S(3,3));

s = det(U)*det(V);

dPrime_PM = d2;

if d1~=d2 && d2~=d3
    ncase = 1;
elseif d1==d2 && d2==d3
    ncase = 3;
else
    ncase = 2;
end

if ncase~=1
    error('HOMOGRAPHY:DECOMPOSE','This motion case is not implemented.');
end

% we deal only with case 1, which implies d1~=d3
% see Faugeras paper Eq. 12:
x1_PM = sqrt((d1*d1 - d2*d2)/(d1*d1 - d3*d3));
x2 = 0;
x3_PM = sqrt((d2*d2 - d3*d3)/(d1*d1 - d3*d3));

e1 = [1 -1  1 -1];
e3 = [1  1 -1 -1];

decomp = struct('R',{},'t',{},'n',{},'d',{},'Rs',{},'ts',{});

% solution counter
sCnt = 0;

% Case 1: d'>0
for signs=1:4
    sCnt = sCnt+1;
    
    decomp(sCnt).d = s*dPrime_PM;
    
    % eq 13
    dSinTheta = (d1 - d3) * x1_PM * x3_PM * e1(signs) * e3(signs) / d2;
    dCosTheta = (d1 * x3_PM * x3_PM + d3 * x1_PM * x1_PM) / d2;
    decomp(sCnt).Rs = [dCosTheta 0 -dSinTheta; 0 1 0; dSinTheta 0 dCosTheta];
    
    % eq 14
    decomp(sCnt).ts = (d1 - d3)* [ x1_PM * e1(signs); 0; -x3_PM * e3(signs)];
    
    ns = [x1_PM * e1(signs);  x2; x3_PM * e3(signs)];
    decomp(sCnt).n = V*ns;
    decomp(sCnt).R = s * U * decomp(sCnt).Rs * V';
    decomp(sCnt).t = U * decomp(sCnt).ts;
end

% Case 1: d'<0
for signs=1:4
    sCnt = sCnt+1;
    
    decomp(sCnt).d = s*-dPrime_PM;
    
    % eq 15
    dSinPhi = (d1 + d3) * x1_PM * x3_PM * e1(signs) * e3(signs) / d2;
    dCosPhi = (d3 * x1_PM * x1_PM - d1 * x3_PM * x3_PM) / d2;
    decomp(sCnt).Rs = [dCosPhi 0 dSinPhi; 0 -1 0; dSinPhi 0 -dCosPhi];
    
    % eq 16
    decomp(sCnt).ts = (d1 + d3) * [x1_PM * e1(signs) ; 0; x3_PM * e3(signs)];
    
    ns = [x1_PM * e1(signs);  x2; x3_PM * e3(signs)];
    decomp(sCnt).n = V*ns;
    decomp(sCnt).R = s * U * decomp(sCnt).Rs * V';
    decomp(sCnt).t = U * decomp(sCnt).ts;
    
end
