% FUNDAMENTALMATRIX - Calculates the fundamental matrix
%
% Usage:    [F, inlyingMatches] = fundamentalmatrix(m1, m2, varargin)
%           [F, inliers]        = fundamentalmatrix(m1, m2, 'getInlierIndex', true, varargin)
%
% Arguments:
%       m1 - 2xN or 3xN set of points in firsts image.
%       m2 - 2xN or 3xN set of points in second image, such that m1<->m2.
%
% Optional arguments:
%      fundmatrixsolution
%           8point - use 8-point algorithm to calculate fundamental matrix
%           7point - use 7-point algorithm to calculate fundamental matrix
%           affine - use affine algorithm to calculate fundamental matrix
%      verbose
%           true  - verbose output 
%           false - no verbose output (default)
%      definiteMatches
%           1x2 cell array, containing 2xN sets of points, that are 
%           definite matches in both images
%      getInlierIndex
%           true  - returns index array of inliers
%           false - returns 2x1 cell array of inlying matches (default)
%
% Returns:
%       F              - The 3x3 fundamental matrix such that x2'Fx1 = 0.
%       inlyingMatches - 2x1 cell array of inlying matches
%       inliers        - array of indices of inlying matches
%
% See Also: RANSACFITFUNDMATRIX
% 
% @author Hanno Jaspers
% @date 2011
function [ F, argout ] = fundamentalmatrix( m1, m2, varargin)

[fundmatrixsolution definiteMatches params] = parse_arguments(varargin);

% assemble homogeneous feature coordinates for fitting of the
% fundamental matrix, note that [x,y] corresponds to [col, row]
x1 = [m1(1,:); m1(2,:); ones(1,size(m1,2))];
x2 = [m2(1,:); m2(2,:); ones(1,size(m1,2))];

% extend x1 and x2 with the definiteMatches FIXME: MAKE SURE THEY ARE
% ALWAYS INLIERS!!
if ~isempty(definiteMatches)
    x1 = [x1 definiteMatches{1}];
    x2 = [x2 definiteMatches{2}];
end
global ransacDistThresh

if isempty(ransacDistThresh)
    t = .001;  % distance threshold for deciding outliers
else
    t = ransacDistThresh;
end

ransacSuccess = false;
while ~ransacSuccess
    try
        switch fundmatrixsolution
            case '7point'
                [F, inliers] = ransacfitfundmatrix7(x1, x2, t, 0);
            case 'affine'
                [F, inliers] = ransacfitaffinefund(x1, x2, t, 0);
            otherwise
                [F, inliers] = ransacfitfundmatrix(x1, x2, t, 0);
                %F = fundmatrix(x1(:,inliers), x2(:,inliers));
        end
        ransacSuccess = true;
    catch ME
        fprintf('Error "%s%" occured while computing fundamental matrix... Repeating step!\n', ME.message);
    end
end

% only at most 8 matches are necessary to compute the fundamental matrix.
% the more accurate those matches are, the more accurate is the funamental
% matrix. we try to find the 8 points that, combined with the definite
% matches produce a fundamental matrix that is the closest to the centroid
% of all the producable fundamental matrices
%for i=1:4:length(inliers)

useInliers = max([25,floor(0.5*length(inliers))]);

if length(inliers)>useInliers
    
    cnt = 200;
    cntPts = useInliers; % 12
    cntDefMatch = length(definiteMatches{1});
    Fs = zeros(3,3,cnt);
    x1i = zeros(3,cntPts+cntDefMatch,cnt);
    x2i = zeros(3,cntPts+cntDefMatch,cnt);
    absErrors = zeros(1,cnt);
    chosenInliers = zeros(cnt,cntPts);
    for i=1:cnt
        inliersPerm = randperm(length(inliers));
        chosenInliers(i,:) = inliers(inliersPerm(1:cntPts));
        x1i(:,:,i) = [ x1(:,chosenInliers(i,:)) definiteMatches{1} ];
        x2i(:,:,i) = [ x2(:,chosenInliers(i,:)) definiteMatches{2} ];
        [Fs(:,:,i)] = fundmatrix(x1i(:,:,i),x2i(:,:,i));
        absErrors(i) = sum(abs(diag(x2i(:,:,i)'*Fs(:,:,i)*x1i(:,:,i))));
    end
    
    % column wise the components of the fundmatrices
    %FFs = reshape(Fs,9,cnt)'; % from 3x3xN to N*9
    %centroid = sum(FFs)/size(FFs,1); % 1x9
    %FFs_norm = FFs - repmat(centroid,cnt,1); % Nx9
    % find the fundmatrix with the minimum euclidean distance to the centroid
    %[minv mini] = min(sum(FFs_norm.*FFs_norm,2)); % just the square dist
    [minv mini] = min(absErrors);
    F = Fs(:,:,mini);
    
    inlyingMatches = cell(2,1);
    inlyingMatches{1} = x1i(1:2,:,mini);
    inlyingMatches{2} = x2i(1:2,:,mini);
    inliers = chosenInliers(mini,:);
else
    inlyingMatches = cell(2,1);
    inlyingMatches{1} = x1(1:2,inliers);
    inlyingMatches{2} = x2(1:2,inliers);
end

if params.getInlierIndex
    argout = inliers;
else
    argout = inlyingMatches;
end

end


function [fundmatrixsolution definiteMatches params] = parse_arguments(args)
nargs = size(args,2);
params = struct();
params.verbose = false;
params.getInlierIndex = false;
fundmatrixsolution = '8point';
definiteMatches = cell(1,2);
if ~isempty(args)
    for i=1:2:nargs
        switch (args{i})
            case 'verbose'
                params.verbose = args{i+1};
            case 'getInlierIndex'
                params.getInlierIndex = args{i+1};
            case 'fundmatrixSolution'
                if sum(strcmpi(args{i+1},{'8point','7point','affine'}))>0
                    params.fundmatrixsolution = args{i+1};
                end
            case 'definiteMatches'
                definiteMatches = args{i+1};
                if size(definiteMatches{1},1) == 2
                    definiteMatches{1} = [definiteMatches{1}; ones(1,size(definiteMatches{1},2))];
                end
                if size(definiteMatches{2},1) == 2
                    definiteMatches{2} = [definiteMatches{2}; ones(1,size(definiteMatches{2},2))];
                end
        end
    end
end
end