% GETPOSEFROMFUNDMATRIX Computes relative pose from a fundamental matrix
% 
% Usage:    [ C, R ] = getPoseFromFundmatrix(F, K1, K2, x1, x2)
%           [ C, R ] = getPoseFromFundmatrix(F, K1, K2, x1, x2, optimize)
%
% Arguments:
%       F  - Fundamental matrix
%       K1 - (intrinsic) calibration matrix of first camera
%       K2 - (intrinsic) calibration matrix of second camera
%       x1 - point correspondences in first camera image (2xN or 3xN)
%       x2 - point correspondences in second camera image (2xN or 3xN)
%       
% Optional arguments:
%       optimize - 0/1 additional optimization of pose (default: 0)
%
% Returns:
%       P (camera matrix) instead of:
%       C - position (translation vector) with |C| = 1
%       R - rotation matrix
%
% See Also: posesFromEssentialMatrix, poseOptimization
%
% @author Hanno Jaspers
% @date 2011
function [ P ] = getPoseFromFundmatrix( F, K1, K2, x1, x2, optimize)

    if nargin < 5
        error('Not enoug arguments!');
    end
    
    if nargin < 6
        optimize = 0;
    end

    % create essential Matrix
    E = K2'*F*K1;
    
    % get rotation and translation from essential matrix
    [Pcell] = posesFromEssentialMatrix(E);
    
    % resolve ambiguity of the four solutions
    [P correctSolution] = cheiralityConstraint(Pcell, K1, K2, x1, x2);
    
    [C R] = getCRfromP(P);
    
    % optimize result
    if optimize
        [C R] = poseOptimization(x1, x2, R, C, K1, K2);
    end
    
    P = getPfromCR(C,R);

end

