% GETPOSEFROMHOMOGRAPHY Calculate pose from homography.
%
% Usage: [ R t ] = getPoseFromHomography( H, K, imgSize, realSize)
%
% Arguments:
%       H - homography matrix
%       K - (intrinsic) camera calibration matrix
%       imgSize  - object image size in px
%       realSize - real object/image dimensions in a specific unit, eg. mm
%
% Returns:
%       R - rotation matrix of object's rotation in relation to viewpoint
%       t - translation vector of upper left image corner in relation to
%           the viewpoint; values are in the specific unit (eg. mm)
%
% This function calculates the rotation and translation of a reference
% object in relation to the current viewpoint. Since the absolute
% dimensions of the object are known, the translation is given in the
% specific units as passed in the argument realSize.
%
% See Also: HOMOGRAPHY
%
% @author Hanno Jaspers
% @date 2011

function [ R t_cm ] = getPoseFromHomography( H, K, imgSize, realSize)

lambda = (1/norm(K\H(:,1))+1/norm(K\H(:,2)))/2;
t = lambda * (K\H(:,3));

% compare diagonals
absScale = norm(realSize)/(norm(imgSize)); % px units to cm
t_cm = absScale * t;

r1 = lambda * (K\H(:,1));
r2 = lambda * (K\H(:,2));
r3 = cross(r1,r2);

R = [r1 r2 r3];

% estimate best rotation matrix from matrix R (does generally not satisfy
% properties of a rotation matrix)
[U S V] = svd(R);
R = U*V';

end

