% GLOBALPOSE Calculates global pose of a camera image.
%
% Usage:    [ P, R, q] = globalpose( P1, P2, PR, K1, K2, RS )
%
% Arguments:
%       P1 - path of first image
%       P2 - path of second image
%       RP - path of reference image
%       K1 - calibration matrix of first camera
%       K2 - calibration matrix of second camera
%       RS - absolute size of reference image (2x1 vector)
%
% Optional arguments:
%       verbose - true/false
%       extrinsic - 3x4 Matrix of extr. parameters of first cam
%       scale - 1x3 vector of scaling factors for images
%       repetitions - number of repetitions of algorithm
%       reconstructPoints - if true, reconstructed points will be in q.X
%
% Returns:
%       P - Position of second image
%       R - Rotation of second image
%       q - quality value of result (sight-ray difference) - struct
%       Note: if repetitions > 1, then the returned fields are cell arrays
% 
% Calculates global pose of a camera image using a calibration
% object with known size.
%
% See Also: SETUP, HOMOGRAPHY, POSESFROMESSENTIALMATRIX
%
% @author Hanno Jaspers
% @date 2011

function [ BestCamPosition , BestCamRotation, err] = globalpose(varargin)

%% parameter parsing
params = parse_parameters(varargin);

%% further initialization
[K1 K2 img1 img2 refimg] = initialize(params);
im1 = rgb2gray(img1);
im2 = rgb2gray(img2);
refim = rgb2gray(refimg);

%% calculate sift features in all images
if params.verbose
    tic
end

[m1, m2, FRAMES1, DESC1, FRAMES2, DESC2] = getPointCorrespondencesWithSift(im1, im2);

[FRAMES3,DESC3] = loadSift(refim);

global match_thresh_reference;

MATCHES1 = vl_ubcmatch(DESC3,DESC1,match_thresh_reference);
MATCHES2 = vl_ubcmatch(DESC3,DESC2,match_thresh_reference);


if params.verbose
%     figure(1);
%     clf;
%     vertsize = max(size(im1,1),size(refim,1));
%     tmp1 = refim;
%     tmp2 = im1;
%     
%     if vertsize-size(refim,1) > 0; tmp1 = padarray(tmp1,[vertsize-size(refim,1) 0],0,'post'); end;
%     if vertsize-size(im1,1) > 0; tmp2 = padarray(tmp2,[vertsize-size(im1,1) 0],0,'post'); end;
%     
%     imshow([tmp1 tmp2]);
%     hold on;
%     mt1 = MATCHES1(1,:);
%     mt2 = MATCHES1(2,:);
%     X = [FRAMES3(1,mt1); FRAMES3(2,mt1)]; % model points [x y]'
%     U = [FRAMES1(1,mt2); FRAMES1(2,mt2)]; % image points [u v]'
%     for i = 1:length(MATCHES1)
%         line([X(1,i) U(1,i)+size(refim,2)], [X(2,i) U(2,i)],'color',[0 0 1]);
%     end
%     scatter(X(1,:),X(2,:));
%     scatter(U(1,:)+size(refim,2),U(2,:));
%     hold off;
    fprintf('Time to compute and match sift-features in all three images: %.4f\n', toc);
end

if params.repetitions > 1
   resultC = cell(1,params.repetitions);
   resultR = cell(1,params.repetitions);
   resultErr = cell(1,params.repetitions);
end

for iteration = 1:params.repetitions
    
    fprintf('%i... ',iteration);
    
    %% Calculation of homography between refimg and img1/img2
    if params.verbose 
        tic
    end
    
    counter = 0;
    H1 = [];
    H2 = [];
    
    math_thresh_ref_bkp = match_thresh_reference;
    while isempty(H1) && counter < 10
        [ H1 ] = homography(FRAMES3, DESC3, FRAMES1, DESC1, MATCHES1);
        match_thresh_reference = match_thresh_reference+0.1;
        counter = counter + 1;
    end
    
    match_thresh_reference = math_thresh_ref_bkp;
    counter = 0;
    
    while isempty(H2) && counter < 10
        [ H2 ] = homography(FRAMES3, DESC3, FRAMES2, DESC2, MATCHES2);
        match_thresh_reference = match_thresh_reference+0.1;
        counter = counter+1;
    end
    match_thresh_reference = math_thresh_ref_bkp;
    
    if params.verbose
        fprintf('Time to compute homographies for the two image pairs: %.4f\n', toc);
    end

    stopimmediately = 0;
    if isempty(H1)
        stopimmediately = 1;
        warning('HOMOGRAPHY:empty','The reference object was not found in image 1. The homography could not be calculated. Returning -1 -1 -1!');
    end
    if isempty(H2)
        stopimmediately = 1;
        warning('HOMOGRAPHY:empty','The reference object was not found in image 2. The homography could not be calculated. Returning -1 -1 -1!');
    end
    if stopimmediately == 1
        BestCamPosition = -1;
        BestCamRotation = -1;
        err = -1;
        if params.repetitions > 1
            resultC{iteration} = BestCamPosition;
            resultR{iteration} = BestCamRotation;
            resultErr{iteration} = err;
        end
        continue;
    end
    
    % calculate the four corner points of the reference image in the two imgs
    cornerPtsRef = [0 0 1; 0 size(refim,1) 1; size(refim,2) size(refim,1) 1; size(refim,2) 0 1]';
    cornerPtsIm1 = H1*cornerPtsRef;
    cornerPtsIm1  = cornerPtsIm1 ./ repmat(cornerPtsIm1(3,:),3,1);
    cornerPtsIm2 = H2*cornerPtsRef;
    cornerPtsIm2  = cornerPtsIm2 ./ repmat(cornerPtsIm2(3,:),3,1);
    
%     cpr2 = [size(refim,2)/2 size(refim,1)/2 1; 0 size(refim,1)/2 1; size(refim,2)/2 size(refim,1) 1; size(refim,2) size(refim,1)/2 1; size(refim,2)/2 0 1]';
%     cpr2 = [cpr2 [rand(1,20)*size(refim,2); rand(1,20)*size(refim,1); ones(1,20)]]; 
%     cpi1 = H1*cpr2;
%     cpi1  = cpi1 ./ repmat(cpi1(3,:),3,1);
%     cpi2 = H2*cpr2;
%     cpi2  = cpi2 ./ repmat(cpi2(3,:),3,1);
%     m1 = cpi1(1:2,:);
%     m2 = cpi2(1:2,:);
    definiteMatches = cell(1,2);
    definiteMatches{1} = cornerPtsIm1;
    definiteMatches{2} = cornerPtsIm2;

    [R_h1 t_cm1] = getPoseFromHomography(H1,K1,size(refim),params.refImgSize);
    [R_h2 t_cm2] = getPoseFromHomography(H2,K2,size(refim),params.refImgSize);

    % Homography from image plane 1 to image plane 2 is:
%     H3 = H2/H1;
    % remove intrinsic calibration parameters, so the Ks will be eye(3)
%     H3 = matnorm(K2\H3*K1);
    % --> decompose
%     [decomp] = decomposeHomography(H3);
%     [R t n] = chooseCorrectHomographyDecomposition(H3,decomp,K1\cornerPtsIm1,K2\cornerPtsIm2);
    % re-apply?
    % H3 = matnorm(K2*H3/K1)
%     RH = R;
%     TH = t;
    
    %% undistort point correspondences
%     definiteMatches{1} = normalize_pixel(definiteMatches{1}, [K1(1,1); K1(2,2)], K1(1:2,3), kc1, K1(1,2),params.scale);
%     definiteMatches{2} = normalize_pixel(definiteMatches{2}, [K2(1,1); K2(2,2)], K2(1:2,3), kc2, K2(1,2),params.scale);
%     m1_n = normalize_pixel(m1, [K1(1,1); K1(2,2)], K1(1:2,3), kc1, K1(1,2));
%     m2_n = normalize_pixel(m2, [K2(1,1); K2(2,2)], K2(1:2,3), kc2, K2(1,2));
    
    %% Calculation of fundamental matrices based on matches between im1 and im2
    [F inlyingMatches] = fundamentalmatrix(m1, m2, ...
                           'fundmatrixsolution', params.fundmatrixsolution,...
                           'definiteMatches', definiteMatches);
%     [F inlyingMatches] = fundamentalmatrix(m1_n, m2_n, ...
%                            'fundmatrixsolution', params.fundmatrixsolution);
    x1i = inlyingMatches{1};
    x2i = inlyingMatches{2};

    %% calculate 4 relative poses from essential matrix
    E = K2'*F*K1;
    %[Pcell] = posesFromEssentialMatrixHorn(E, params.verbose);
    [Pcell] = posesFromEssentialMatrix(E, params.verbose);
    %Ps = getCameraMatrix(E);

    %% resolve ambiguity of the four solutions
    %[P, correctSolution] = cheiralityConstraint(Pcell, K1, K2, x1i, x2i);
    [P, votes] = getCorrectCameraMatrix(cat(3,Pcell{:}), repmat(E,[1,1,4]), K1, K2, [x1i; ones(1,size(x1i,2)); x2i; ones(1,size(x1i,2))]);
    [maxvotes, correctSolution] = max(votes);

    [C R] = getCRfromP(P);
    
%     if params.verbose
%         [Ccell Rcell] = getCRfromP(Pcell);
%         X = reconstructPoint(x1i,x2i,eye(3),[0 0 0]',R,C,K1,K2);
%         visualizeRelativeSolution(Rcell,Ccell,X,correctSolution,2);
%         
%         %visualizeReprojection(X, x1i, [eye(3) [0;0;0]], K, 4)
%         visualizeReprojection(x2i, X, P, K2, 4)
%         
%     end
    

    %% Optimize result
    [C R err] = poseOptimization(x1i, x2i, R, C, K1, K2, 1);
    
    %C = inv(RT2)*RT1*[0;0;0;1];
    %C = C./repmat(C(4),4,1);
    %C = C(1:3);
    %R = R_h2'*R_h1;

    if params.verbose
        X = reconstructPoint(x1i,x2i,eye(3),[0 0 0]',R,C,K1,K2);
        visualizeReprojection(x2i, X, getPfromCR(C,R), K2, 5)
        imc = [img1 img2];
        figure(1); clf; imshow(imc); hold on;
        for i = 1:length(x1i)
            line([x1i(1,i) x2i(1,i)+size(im1,2)], [x1i(2,i) x2i(2,i)],'color',[0 0 1]);
            %line([m1(2,n) m2(2,n)+size(im1,2)], [m1(1,n) m2(1,n)],'color',[0 0 1])
        end
        line([cornerPtsIm1(1,:) cornerPtsIm1(1,1)], ...
            [cornerPtsIm1(2,:) cornerPtsIm1(2,1)],...
            'Color','r','LineWidth',2);
        line([cornerPtsIm2(1,:)+repmat(size(im1,2),1,4) cornerPtsIm2(1,1)+size(im1,2)], ...
            [cornerPtsIm2(2,:) cornerPtsIm2(2,1)],...
            'Color','r','LineWidth',2);
        plot(x1i(1,:),x1i(2,:),'r+');
        plot(x2i(1,:)+size(im1,2),x2i(2,:),'g+');
        hold off
    end

    %% Calc metric relative pose in cm
    % (C1C2) = (C1X)+(XC2)
    X = reconstructPoint(definiteMatches{1}(:,1),definiteMatches{2}(:,1),eye(3),[0 0 0]',R,C,K1,K2);
    k1 = norm(X);
    k2 = norm(C-X);
    l1 = norm(t_cm1);
    l2 = norm(t_cm2);

    s1 = l1/k1;
    s2 = l2/k2;

    % 4 possible solutions:
    % we have 2 scale factors s1 and s2 received from the homographies, which
    % should be equal but differ in real images.
    % Solution 1: C_hat = s1*X + s2*(C-X)
    % Solution 2: C_hat = s1*C
    % Solution 3: C_hat = s2*C
    % Solution 4: C_hat = (s1+s2)/2*C

    % Camera positions
    CamPositionsPre = ones(4,4);
    CamPositionsPre(1:3,1) = s1*X + s2*(C-X);
    CamPositionsPre(1:3,2) = s1*C;
    CamPositionsPre(1:3,3) = s2*C;
    CamPositionsPre(1:3,4) = (s1+s2)/2*C;

    [initialPos initialRot] = getCRfromP(params.extrinsic);
    %initialPos = [0 0 0]';
    %initialRot = eye(3);
    
    % apply initial extrinsic parameters:
    CamPositions = [initialRot' initialPos]*CamPositionsPre;

    % choose the first value as the camera position
    CamPosition = CamPositions(:,params.camPosition);%mean(CamPositions,2);
    CamRotation = R*initialRot;

    
    %initialRot = params.extrinsic(1:3,1:3)';
    %initialPos = -initialRot*params.extrinsic(:,4);

    % Reconstruct points and get distances of view-lines
    [X dists] = reconstructPoint(x1i,x2i,initialRot,initialPos,CamRotation,CamPosition,K1,K2);
%     [X dists] = reconstructPoint(x1i,x2i,eye(4),[0 0 0]',CamRotation,CamPosition,K1,K2);
%     X = [initialRot initialPos] * [X; ones(1,size(X,2))];
%     X = X/repmat(X(4,:),4,1);
%     X = X(1:3,:);
    
    
    % calculate likelihood of a plausible solution -> distance variance
    CPs_n = CamPositions(:,1:3) - repmat(mean(CamPositions(:,1:3),2),1,3);%CamPositions - repmat(CamPosition,1,4);
    dists = sqrt(sum(CPs_n.*CPs_n));
    meanDist = mean(dists);
    varDist = var(dists-meanDist);

    % "normalize" variance by scale
    varDistScaled = varDist*2/(s1+s2);
    meanDistScaled = meanDist*2/(s1+s2);

    if params.verbose 
       fprintf('Distance of reference object to first camera: %.2f - to second camera: %.2f\n',[l1,l2]);
%         fprintf('(Absolute) ratio of scaling factors:        %8.4f\n',abs(s1/s2));
%         fprintf('Mean distance of position (scaled):         %8.4f %8.4f\n',meanDist,meanDistScaled);
%         fprintf('Variance of distances of position (scaled): %8.4f %8.4f\n',varDist,varDistScaled );
%         fprintf('Mean   distance of reconstruced lines:      %8.4f\n',mean(dists));
%         fprintf('Median distance of reconstruced lines:      %8.4f\n',median(dists));
    end
 
    quality = 1/abs(abs(s1/s2)-1);
    
    BestCamPosition = CamPosition;
    BestCamRotation = CamRotation;
    bestQuality = quality;
    
    if params.verbose
        fh = figure(3);
        clf;
        hold on;
        scale1 = K1(1)/10*norm(params.refImgSize)/norm(size(refim));
        scale2 = K2(1)/10*norm(params.refImgSize)/norm(size(refim));
        drawCamera(initialRot,initialPos,scale1,fh,struct('Color','g'));
        drawCamera(CamRotation,CamPositions(:,1),scale2,fh,struct('Color','r'));
        drawCamera(CamRotation,CamPositions(:,2),scale2,fh,struct('Color','y'));
        drawCamera(CamRotation,CamPositions(:,3),scale2,fh,struct('Color','c'));
        drawCamera(CamRotation,CamPositions(:,4),scale2,fh,struct('Color','m'));
        % ---------------------------------------------------------------------
        % Reconstruct points
        % ---------------------------------------------------------------------
        % first: throw out points, that are probably outliers --> simple
        % distance criterion
        Xm = X-repmat(mean(X,2),1,size(X,2));
        % distance to mean:
        Xdists = sqrt(sum(Xm.^2));
        % throw out those, whose distance is greater than 2*std(Xdists)
        X_red = X(:,Xdists < 2*std(Xdists));
        %X_red = X;
        fprintf('Threw out %d points.\n',length(X)-length(X_red));
        scatter3(X_red(1,:),X_red(2,:),X_red(3,:));
        hold off;
        
        P = getPfromCR(BestCamPosition,BestCamRotation);
        visualizeReprojection(x2i, X, P, K2, 6)
    end

    P = getPfromCR(BestCamPosition,BestCamRotation);
    reprErrs = reprojectionError(x2i,X,P,K2,1);
    err.reprojectionErr = mean(reprErrs);
    err.reprojectionErrMed = median(reprErrs);
    err.quality = bestQuality;
    
    if params.verbose
        fprintf('Mean reprojection error: %.2f px\n',err.reprojectionErr);
    end
    
    if params.getPoints
       err.X = X; 
    end
    
    if params.repetitions > 1
        resultC{iteration} = BestCamPosition;
        resultR{iteration} = BestCamRotation;
        resultErr{iteration} = err;
        % last run
        if iteration==params.repetitions
            C = mean(horzcat(resultC{:}),2);
            R = mean(cat(3,resultR{:}),3);
            R = makeRotationMatrix(R);
            [X dists] = reconstructPoint(x1i,x2i,initialRot,initialPos,R,C,K1,K2);
            resultErr = [resultErr, X];
        end
    end
    
end

fprintf('\n');

if params.repetitions > 1
    BestCamPosition = resultC;
    BestCamRotation = resultR;
    err = resultErr;
end


end




function [params] = parse_parameters(args)
params = struct;
nargs = size(args,2);

% default arguments
params.verbose = false;
params.fundmatrixsolution = '8point';
params.scale = 1;
params.repetitions = 10;
params.getPoints = 0;
params.camPosition = 1;

% not enough arguments...
%if (stdargin<2)
%    error('MATLAB:siftfund - Argument parsing: %s', 'You have to specify two images!');
%end
if isempty(args)
    % default parameters
    params.verbose = true;
%     params.img1path = '../data/2010-12-13_Test_JPEG/curtiz_000.jpg';
%     params.img2path = '../data/2010-12-13_Test_JPEG/fassbinder_000.jpg';
%     params.refImgPath = '../data/2010-12-13_Test_JPEG/reference_finca.jpg';
    params.img1path = 'D:\Documents\Dropbox\Documents\Uni\Paper\Studienarbeit\selectedData\threeview/undist_canon_home_01.jpg';
    params.img2path = 'D:\Documents\Dropbox\Documents\Uni\Paper\Studienarbeit\selectedData\threeview/undist_canon_home_05.jpg';
    params.refImgPath = 'D:\Documents\Uni\SS10\Studienarbeit\repository\codebase/data/graffiti2.jpg';
    params.scale = [1 1 1];
    params.refImgSize = 2.835*[9 7]; % it was a checkerboard pattern before
    R = angle2dcm(0,0,-pi/2);
    t = [0 0 150]';
    params.extrinsic = getPfromCR(t,R);
    
    % For Canon Ixus 80 IS
    fc = [ 1769.211730204996700 ; 1768.074408993500600 ];
    cc = [ 796.465230721117680 ; 614.826367237053320 ];
    params.calibrationMatrices = ...
       [fc(1) 0 cc(1); ...
        0 fc(2) cc(2); ...
        0 0 1];
    
    kc = [ -0.179937187723909 ; 0.137110822128179 ; 0.000814001177183 ; -0.001387955972617 ; 0.000000000000000 ];
    params.distortion = {kc, kc};
else
    %if ~ischar(varargin{1}) 
    params.img1path = args{1};
    params.img2path = args{2};
    params.refImgPath = args{3};
    params.refImgSize = args{6};
    params.K1 = args{4};
    params.K2 = args{5};
    params.extrinsic = eye(3,4);
    % parse optional arguments
    for i=7:2:nargs
        switch (args{i})
            case 'verbose'
                params.verbose = args{i+1};
            case 'fundmatrixSolution'
                if sum(strcmpi(args{i+1},{'8point','7point','affine'}))>0
                    params.fundmatrixsolution = args{i+1};
                end
            case 'calibrationMatrices'
                params.calibrationMatrices = args{i+1};
            case 'scale'
                params.scale = args{i+1};
            case 'extrinsic'
                params.extrinsic = args{i+1};
            case 'repetitions'
                params.repetitions = args{i+1};
            case 'reconstructPoints'
                params.getPoints = args{i+1};
            case 'CamPositionChoice'
                params.camPosition = args{i+1};
        end
    end
end
if length(params.scale) == 1
    params.scale = repmat(params.scale,1,3);
end


end



function [K1 K2 img1 img2 refImg] = initialize(params)

% resize with "scale"
img1 = imresize(imread(params.img1path),params.scale(1));
img2 = imresize(imread(params.img2path),params.scale(2));
refImg = imresize(imread(params.refImgPath),params.scale(3));

if isfield(params,'calibrationMatrices')
    % load calibration matrix
    if iscell(params.calibrationMatrices)
        if length(params.calibrationMatrices)>=2
            K1 = params.calibrationMatrices{1};
            K2 = params.calibrationMatrices{2};
        else
            K1 = params.calibrationMatrices{1};
            K2 = params.calibrationMatrices{1};
        end
    else
        K1 = params.calibrationMatrices;
        K2 = params.calibrationMatrices;
    end
else
    K1 = params.K1;
    K2 = params.K2;
end

K1 = K1*params.scale(1); K1(3,3) = 1;
K2 = K2*params.scale(2); K2(3,3) = 1;
%kc1 = kc1 * params.scale;
%kc2 = kc2 * params.scale;

end

