% HOMOGRAPHY - Calulates the homography
%
% Usage:    H = homography( FRAMES1, DESC1, FRAMES2, DESC2 )
%      
% Arguments:
%       FRAMES1, DESC1 - SIFT features of reference object
%       FRAMES2, DESC2 - SIFT features of complete image
% 
% Returns:
%       H - Homography of transformation from object->image.
%
% See Also: HOMOGRAPHY2D, RANSAC
% 
% @author Peter Kovesi, Hanno Jaspers
% @date 2011
function [ H ] = homography( FRAMES1, DESC1, FRAMES2, DESC2, MATCHES )

global match_thresh_reference
if isempty(match_thresh_reference)
    match_thresh_reference = 1.5;
end

%% match sift features
if nargin < 5
    MATCHES = vl_ubcmatch(DESC1,DESC2,match_thresh_reference);
end
%global scene current_ref current_img
% if ~isempty(current_ref) && ~isempty(current_img)
%     clf;
%     plotmatches(scene.reference_objects(current_ref).imagedata, scene.imagelists(current_img(1)).images(current_img(2)).imagedata, FRAMES1, FRAMES2, MATCHES);
% end
% the indices to the SIFT eeatures for each image
mt1 = MATCHES(1,:);
mt2 = MATCHES(2,:);

% the point coordinates of the matching SIFT points
X = [FRAMES1(1,mt1); FRAMES1(2,mt1)]; % model points [x y]'
U = [FRAMES2(1,mt2); FRAMES2(2,mt2)]; % image points [u v]'

% Undistort points

%H = ransacfithomography(X,U,0.005);


%% Calculate affine transformation parameters and projective transformation

% Find and remove outliers by comparing predicted position using the
% calculated parameters of the affine transformation and the actual
% positions; criteria: RANSAC

if size(X,1) == 2    % extend data with homogeneous scale factor of 1
    X = [X; ones(1,size(X,2))];
    U = [U; ones(1,size(U,2))];
end
% 
% % Normalise each set of points so that the origin is at centroid and
% % mean distance from origin is sqrt(2).  normalise2dpts also ensures the
% % scale parameter is 1.

homographyfn = @homography2d;
%distfn     = @geometricError;
distfn      = @homogdist2d;
%degenfn    = @isdegenerate;
degenfn    = @isdegenerate_coll;

s = 4;    % 4 points needed
t = .005;     % distance threshold for deciding outliers
feedback = 0; % feedback output for ransac
maxDataTrials = 350;
maxTrials = 700;

% normalise points --> no! normalisation is done in homography2d
 [X, T1] = normalise2dpts(X);
 [U, T2] = normalise2dpts(U);

 
% save warning state
warnstate = warning('query', 'all');
%disable all warnings
warning off all
lastwarn('')

% calculate Affinity and Homography
success = false;
%while ~success
    try
        [H, inliersHomogr] = ransac([X; U], homographyfn, distfn, degenfn, s, t, feedback, maxDataTrials, maxTrials);
        success = true;
    catch ME
        fprintf('%s  --  %s\n',ME.identifier, ME.message);
        %fprintf('Error occured while computing homography...!\n');
    end
%end

if ~isempty(lastwarn)
    success = false;
end

% restore warning state
warning(warnstate)

if success

    % Homography inliers
    X_h = X(:,inliersHomogr);
    U_h = U(:,inliersHomogr);

    % Homography calculation (LS fit with inliers)
    H = homography2d(X_h,U_h);
    
    % Denormalise
    H = T2\H*T1;
    
    H = H/H(9);
    
else
    H = [];
    %match_thresh_reference = match_thresh_reference + 0.05
    %H = homography(FRAMES1,DESC1,FRAMES2,DESC2);
end

end




function [bestInliers, M] = geometricError(M,x,t) 
    if iscell(M)
        M = M{1};
    end
    
    X = x(1:3,:);
    U = x(4:6,:);
    
    Xtrans = M*X;
    Xtrans = Xtrans ./ repmat(Xtrans(3,:),3,1);
    
    d = sum((U-Xtrans).*(U-Xtrans));
    
    bestInliers = find(abs(d) < t);     % Indices of inlying points
end


function r = isdegenerate(x)
    if sum(sum(isinf(x)))>0 || sum(sum(isnan(x)))>0
        r = 1;
    else
        r = 0;
    end
end

%----------------------------------------------------------------------
% Function to evaluate the symmetric transfer error of a homography with
% respect to a set of matched points as needed by RANSAC.
% (c) 2004-2005 Peter Kovesi
function [inliers, H] = homogdist2d(H, x, t)
    
    x1 = x(1:3,:);   % Extract x1 and x2 from x
    x2 = x(4:6,:);    
    
    % Calculate, in both directions, the transfered points    
    Hx1    = H*x1;
    invHx2 = H\x2;
    
    % Normalise so that the homogeneous scale parameter for all coordinates
    % is 1.
    
    x1     = hnormalise(x1);
    x2     = hnormalise(x2);     
    Hx1    = hnormalise(Hx1);
    invHx2 = hnormalise(invHx2); 
    
    d2 = sum((x1-invHx2).^2)  + sum((x2-Hx1).^2);
    inliers = find(abs(d2) < t);    
end
    
%----------------------------------------------------------------------
% Function to determine if a set of 4 pairs of matched  points give rise
% to a degeneracy in the calculation of a homography as needed by RANSAC.
% This involves testing whether any 3 of the 4 points in each set is
% colinear. 
% (c) 2004-2005 Peter Kovesi
function r = isdegenerate_coll(x)

    x1 = x(1:3,:);    % Extract x1 and x2 from x
    x2 = x(4:6,:);    
    
    r = ...
    iscolinear(x1(:,1),x1(:,2),x1(:,3)) | ...
    iscolinear(x1(:,1),x1(:,2),x1(:,4)) | ...
    iscolinear(x1(:,1),x1(:,3),x1(:,4)) | ...
    iscolinear(x1(:,2),x1(:,3),x1(:,4)) | ...
    iscolinear(x2(:,1),x2(:,2),x2(:,3)) | ...
    iscolinear(x2(:,1),x2(:,2),x2(:,4)) | ...
    iscolinear(x2(:,1),x2(:,3),x2(:,4)) | ...
    iscolinear(x2(:,2),x2(:,3),x2(:,4));
end