% POSEOPTIMIZATION Optimize relative pose (position+orientation)
% 
% Usage:    [ C, R ] = poseOptimization(m1, m2, R, t, K1, K2, mode = 0)
%
% Arguments:
%       m1 - point correspondences in first camera image
%       m2 - point correspondences in second camera image
%       R  - rotation matrix of second camera
%       t  - translation vector of second camera
%       K1 - (intrinsic) calibration matrix of first camera
%       K2 - (intrinsic) calibration matrix of second camera
%
% Optional arguments:
%       mode - 0 -> reconstruction error (default)
%              1 -> reprojection error
%
% Returns:
%       C - optimized position (translation vector) on unity sphere
%       R - optimized rotation matrix
%       err - struct with err.resnorm and err.residual
%
%   The first camera is supposed to be in [0,0,0] with no rotation.
%
%   The position search-space is the unity-sphere. Starting point
%   is t. The orientation is searched in all 3 dimensions, starting at the
%   angles for R.
%   Thus, the optimization vector is a 1x5 vector: [theta phi rx ry rz]
%   
%   Sphere equations (r = 1): 
%       x = sin(phi)*cos(theta)
%       y = sin(phi)*sin(theta)
%       z = cos(phi)
%           0 <= phi < pi and 0 <= theta < 2pi
%       theta = acos(z)
%       phi   = atan2(y,x)
%
% See Also: RECONSTRUCTPOINT
%
% @author Hanno Jaspers
% @date 2011
function [ C, R, err ] = poseOptimization(m1, m2, R, t, K1, K2, mode)

if nargin < 7
    mode = 0;
end

% t to theta and phi
[theta phi] = sphereXYZ2ThetaPhi(t);

% decompose R into rz, ry, rx
[rz ry rx] = dcm2angle(R);

options = optimset('Display','iter','MaxFunEvals',800,'MaxIter',150,'TolFun',1e-8);

% create a handle to an anonymous function  to cope with the extra params
f1 = @(x)reconstructionError(m1,m2,K1,K2,x);
f2 = @(x)reprojectionErrorWrapper(m1,m2,K1,K2,x);
% f1 = @(x)reconstructionErrorBox(m1,m2,K1,K2,x);
% f2 = @(x)reprojectionErrorWrapperBox(m1,m2,K1,K2,x);

if mode == 1
    f = f2;
else
    f = f1;
end

x0 = [theta phi rx ry rz]; % Starting guess
%x0 = [t(1) t(2) t(3) rx ry rz];

lb = x0 - [ 0.2*pi 0.2*pi 0.4  0.4  0.4];        % lower bounds
ub = x0 + [ 0.2*pi 0.2*pi 0.4  0.4  0.4];        % upper bounds
%lb = x0 - [ 0.5 0.5 0.5 0.4 0.4 0.4];        % lower bounds
%ub = x0 + [ 0.5 0.5 0.5 0.4 0.4 0.4];        % upper bounds

%preOptimizationErr = mean(feval(f,x0));

[x,resnorm,residual] = lsqnonlin(f,x0,lb,ub,options);   % Invoke optimizer
%[x,resnorm] = lsqnonlin(f2,x0,[],[],options);   % Invoke optimizer
%[C,resnorm] = simulannealbnd(f1,x0,lb,ub,options)   % Invoke optimizer

%postOptimizationErr = mean(feval(f,x));

err = struct('residual',residual,'resnorm',resnorm);

%if preOptimizationErr <= postOptimizationErr
%    fprintf('Optimization did not improve pose!\n');
    %R = R,
%    C = t;
%else
    R = angle2dcm(x(5),x(4),x(3));
    C = sphereThetaPhi2XYZ(x(1),x(2));
    %R = angle2dcm(x(6),x(5),x(4));
    %C = x(1:3)';
    %C = C/norm(C);
%end

end

function [dists] = reconstructionError(corr1,corr2,K1,K2,x)
    R = angle2dcm(x(5),x(4),x(3));
    t = sphereThetaPhi2XYZ(x(1),x(2));
    [recPoints dists] = reconstructPoint(corr1,corr2,eye(3),zeros(3,1),R,t,K1,K2);
end

function [d] = reprojectionErrorWrapper(m1,m2,K1,K2,x)

    R = angle2dcm(x(5),x(4),x(3));
    t = sphereThetaPhi2XYZ(x(1),x(2));
    %R = angle2dcm(x(6),x(5),x(4));
    %t = x(1:3)';
    
    [X dists] = reconstructPoint(m1,m2,eye(3),zeros(3,1),R,t,K1,K2);
    
    % pad X with 1
    X = [X; ones(1,size(X,2))];
    
    % camera matrices
    %P1 = [eye(3) [0;0;0]];
    P2 = getPfromCR(t,R);
    
    d = reprojectionError(m2,X,P2,K2,1);
    
end

function [dists] = reconstructionErrorBox(corr1,corr2,K1,K2,x)
    R = angle2dcm(x(6),x(5),x(4));
    t = x(1:3).';
    [recPoints dists] = reconstructPoint(corr1,corr2,eye(3),zeros(3,1),R,t,K1,K2);
end

function [d] = reprojectionErrorWrapperBox(m1,m2,K1,K2,x)

    R = angle2dcm(x(6),x(5),x(4));
    t = x(1:3).';
    
    [X dists] = reconstructPoint(m1,m2,eye(3),zeros(3,1),R,t,K1,K2);
    
    % pad X with 1
    X = [X; ones(1,size(X,2))];
    
    % camera matrices
    P2 = getPfromCR(t,R);
    
    d = reprojectionError(m2,X,P2,K2,1);
    
end
