% POSEFROMESSENTIALMATRIX Calculates relative poses from the essential matrix.
%
% Usage: [ R, C ] = posesFromEssentialMatrix( E, verbose=false )
%
% Arguments:
%       E - 3x3 essential matrix
%
% Optional arguments:
%       verbose - verbose output
%
% Returns:
%       P - 1x4 cell array of possible camera matrices [R t]
%
% Calculates four possible solutions for rotation and translation of the
% camera. Based on Hartley/Zisserman, "Multiple View Geometry 2nd", p. 258.
%
% See Also: CHEIRALITYCONSTRAINT
%
% @author Hanno Jaspers
% @date 2011
function [ P ] = posesFromEssentialMatrix( E, verbose)

if nargin < 2
    verbose = false;
end

%% Calculation of relative pose
% Essential matrix equals fundamental matrix, if the camera matrices
% are identity matrices -> E = P2'*F*P1 //// K'*F*K
%P1 = eye(3,4);
%P2 = eye(3,4);
%E = P2'*F*P1;
%E = E(1:3,1:3);

[U S V] = svd(E);
% calculate average of singular values to compensate for inexact data -----
S1 = (S(1,1)+S(2,2))/2;

E_new = U * diag([S1, S1, 0]) * V';

% scale essential matrix, so that the singular values are 1 ---------------
E_new = E_new/S1;

[U S V] = svd(E_new);
%%
% S should be diag(a1,a2,a3) with a1=a2 and a3=0...
% since that's seldom the case with real data, use the following:
%S = diag([1,1,0]);
W = [0 -1 0; 1 0 0; 0 0 1];
%Z = [0 -1 0; 1 0 0; 0 0 0];

% the four different solutions: -------------------------------------------
% t ~ last column of U
% t = -R*C => C = -inv(R)*t = -R'*t

R1 = U*W*V';
R2 = U*W'*V';
t1 = U(:,3);
t2 = -U(:,3);

if det(R1) < 0
    R1 = -R1;
end

if det(R2) < 0
    R2 = -R2;
end

P = cell(1,4);
P{1} = [R1 t1];
P{2} = [R1 t2];
P{3} = [R2 t1];
P{4} = [R2 t2];

R = cell(1,4);
C = cell(1,4);

R{1} = R1; 
C{1} = -R1' * t1;

R{2} = R1; 
C{2} = -R1' * t2;

R{3} = R2; 
C{3} = -R2' * t1;

R{4} = R2; 
C{4} = -R2' * t2;

% R{1} = U*W*V';
% t{1} = U*[0 0 1]'; 
% C{1} = -R{1}\t{1};
% 
% R{2} = U*W*V';
% t{2} = -U*[0 0 1]'; 
% C{2} = -R{2}\t{2};
% 
% R{3} = U*W'*V';
% t{3} = U*[0 0 1]'; 
% C{3} = -R{3}\t{3};
% 
% R{4} = U*W'*V';
% t{4} = -U*[0 0 1]'; 
% C{4} = -R{4}\t{4};

if verbose
    for i=1:4
        fprintf('---------------------------------------------\n');
        fprintf('Solution %i:\n',i);
        fprintf('C: x=%3.5f y=%3.5f z=%3.5f\n',C{i});
        [rz ry rx] = dcm2angle(R{i});
        fprintf('R: x=%3.5f y=%3.5f z=%3.5f\n',rad2deg(rx),rad2deg(ry),rad2deg(rz));
    end
    fprintf('---------------------------------------------\n');
end

end

