% POSESFROMESSENTIALMATRIXHORN Calculates relative poses from the essential matrix.
%
% Usage: [ R, C ] = posesFromEssentialMatrixHorn( E, verbose=false )
%
% Arguments:
%       E - 3x3 essential matrix
%
% Optional arguments:
%       verbose - verbose output
%
% Returns:
%       P - 1x4 cell array of possible camera matrices
%       %%R - 1x4 cell array of possible rotation matrices
%       %%C - 1x4 cell array of possible camera centers/translations
%
% Calculates four possible solutions for rotation and translation of the
% camera. Based on Hartley/Zisserman, "Multiple View Geometry 2nd", p. 258.
%
% See Also: POSESFROMESSENTIALMATRIX, CHEIRALITYCONSTRAINT
function [ P ] = posesFromEssentialMatrixHorn( E, verbose)

if nargin < 2
    verbose = false;
end

% Obtain bxbT
    bbt = 0.5*trace(E*E')*eye(3,3)-E*E';
        
    % We obtain b by selecting the largest row and dividing by the square
    % root of the diagonal. This is for numerical accuracy.
    if (bbt(1,1)>=bbt(2,2) && bbt(1,1)>=bbt(3,3))
        b1 = bbt(1,:)'./sqrt(bbt(1,1));
    elseif (bbt(2,2)>=bbt(1,1) && bbt(2,2)>=bbt(3,3))
        b1 = bbt(2,:)'./sqrt(bbt(2,2));
    elseif (bbt(3,3)>=bbt(1,1) && bbt(3,3)>=bbt(2,2))
        b1 = bbt(3,:)'./sqrt(bbt(3,3));
    end;
    b2 = -b1;
    
    % Compute cofactors of E
    cofE = [cross(E(:,2),E(:,3)),cross(E(:,3),E(:,1)),cross(E(:,1),E(:,2))]';
    
    % Compute two matricex B
    B1 = [ 0      , -b1(3) ,  b1(2);
           b1(3)  ,  0     , -b1(1);
          -b1(2)  ,  b1(1) ,  0    ];
      
    B2 = [ 0      , -b2(3) ,  b2(2);
           b2(3)  ,  0     , -b2(1);
          -b2(2)  ,  b2(1) ,  0    ];

    b1dot = dot(b1,b1);
    b2dot = dot(b2,b2);
      
    % Compute both R
    R1 = b1dot\(cofE'-B1*E);
    R2 = b2dot\(cofE'-B2*E);
    
    % Scale baseline
    b1 = b1./max(abs(b1));
    b2 = b2./max(abs(b2));
    
    
    % make sure, that r1 and r2 are rotation matrices
    [U1 S1 V1] = svd(R1);
    [U2 S2 V2] = svd(R2);
    R1 = U1*V1.';
    R2 = U2*V2.';
    
    
P = cell(1,4);
P{1} = [R1 b1];
P{2} = [R1 b2];
P{3} = [R2 b1];
P{4} = [R2 b2];


R = cell(1,4);
C = cell(1,4);

R{1} = R1; 
C{1} = -R1' * b1;

R{2} = R1; 
C{2} = -R1' * b2;

R{3} = R2; 
C{3} = -R2' * b1;

R{4} = R2; 
C{4} = -R2' * b2;

% P = zeros(3,4,4);
% P(:,:,1) = [R1 b1];
% P(:,:,2) = [R1 b2];
% P(:,:,3) = [R2 b1];
% P(:,:,4) = [R2 b2];

if verbose
    for i=1:4
        fprintf('---------------------------------------------\n');
        fprintf('Solution %i:\n',i);
        fprintf('C: x=%3.5f y=%3.5f z=%3.5f\n',C{i});
        [rz ry rx] = dcm2angle(R{i});
        fprintf('R: x=%3.5f y=%3.5f z=%3.5f\n',rad2deg(rx),rad2deg(ry),rad2deg(rz));
    end
    fprintf('---------------------------------------------\n');
end

end

