% RECONSTRUCTPOINT Reconstructs one or several 3D points.
%
% Usage: [X, dists] = reconstructPoint(m1, m2, R1, t1, R2, t2, K1, K2)
%
% Arguments:
%       m1 - point correspondence(s) in first image, 2x1 or 2xN 
%       m2 - point correspondence(s) in second image, 2x1 or 2xN
%       R1 - rotation matrix of first camera
%       t1 - (homogenous) translation vector of first camera
%       R2 - rotation matrix of second camera
%       t2 - (homogenous) translation vector of second camera
%       K1 - calibration matrix of first camera
%       K2 - calibration matrix of second camera
%
% Returns:
%       X     - reconstructed 3D points (3xN)
%       dists - distance of sight rays
%
% This function reconstructs a 3D point given the point correspondences in
% two views (2d points) and the camera matrices (rotation and translation 
% of the two cameras).
%
% Since the rays won't always intersect, the reconstructed 3D point is
% chosen as the average point between the line of minimum distance between
% the rays.
%
% @author Hanno Jaspers
% @date 2011

% the line representations are given in parametric form:
% line 1: t1 + l*(R1*p1)
% line 2: t2 + m*(R2*p2)
% Ansatz fr Abstand zweier Geraden (wahrscheinlich, aber nicht
% notwendigerweise windschief):
%   geg: g1(l) = A+l*B, g(m) = C+m*D
%   1. Berechne Vektor, der auf beide Richtungsvektoren senkrecht steht /
%      berechne Kreuzprodukt V = B x D (wenn dies 0 ist, dann parallel...)
%   2. Erzeuge Ebene mit Richtungsvektor der ersten Gerade und Vektor aus
%      (1). Diese Ebene schneidet Gerade 2.
%   3. Berechne Schnittpunkt der Ebene mit Gerade 2.
%      A+l*B+r*V = C+m*D   -> lse GLS nach r
%   4. Entfernung = |r*V|
function [ x  dists ] = reconstructPoint( corr1, corr2, R1, t1, R2, t2, K1, K2)

    if size(t1,1) == 1
        t1 = t1.';
    end
    if size(t2,1) == 1
        t2 = t2.';
    end

%     f = [K1(1,1) K2(1,1)];
%     cc = [K1(:,3) K2(:,3)];

    N = size(corr1,2);
    x = zeros(3,N);
    dists = zeros(1,N);
    
    if size(corr1,1)==3; p1 = corr1;
    else p1 = [corr1(1:2,:); ones(1,N)];
    end 
    if size(corr2,1)==3; p2 = corr2;
    else p2 = [corr2(1:2,:); ones(1,N)];
    end
    
    p1 = (K1*R1) \ p1;
    p2 = (K2*R2) \ p2;

    %bs = R1'*p1;
    %ds = R2'*p2;
    bs = p1;
    ds = p2;
    
    for i=1:N
        
        b = bs(:,i);
        d = ds(:,i);
        v = cross(b,d);
        v = v/norm(v); % normalize v to 1 -> unnecessary with f=1 ?

        % coeffs are: [l m r]
        coeffs = [b -d v]\(t2-t1);

        x(:,i) = t1 + coeffs(1)*b + 1/2*coeffs(3)*v;
        dists(i) = norm(coeffs(3)*v);
        
    end


end

    