% RELATIVEPOSE Calculates relative pose of camera image
%
% Usage:    [ C, R, q] = relativepose( IP1, IP2 )
%
% Arguments:
%       IP1 - path of first image
%       IP2 - path of second image
%
% Optional arguments:
%       verbose - true/false
%       extrinsic - 3x4 Matrix of extr. parameters of first cam
%       calibrationMatrices - 1x2 cell array of calibration matrices
%       scale - scaling of images
%
% Returns:
%       C - Position of second image
%       R - Rotation of second image
%       q - quality value of result (sight-ray difference)
% 
% Calculates relative pose of a camera image.
%
% See Also: SETUP, HOMOGRAPHY, POSESFROMESSENTIALMATRIX
%
% @author Hanno Jaspers
% @date 2011

function [ BestCamPosition , BestCamRotation, err] = relativepose(varargin)

%% parameter parsing
params = parse_parameters(varargin);

%% further initialization
[K1 K2 kc1 kc2 img1 img2] = initialize(params);
im1 = rgb2gray(img1);
im2 = rgb2gray(img2);

%% calculate sift features in all images
if params.verbose
    tic
end

[m1, m2, F1, D1, F2, D2] = getPointCorrespondencesWithSift(im1, im2);

if params.verbose
    fprintf('Time to compute sift-features in both images: %.4f\n', toc);
end

if params.repetitions > 1
   resultC = cell(1,params.repetitions);
   resultR = cell(1,params.repetitions);
   resultErr = cell(1,params.repetitions);
end

for iterationCounter = 1:params.repetitions
    fprintf('%i...',iterationCounter);

    %% Calculation of fundamental matrices based on matches between im1 and im2
    [F inlyingMatches] = fundamentalmatrix(m1, m2, ...
                           'fundmatrixsolution', params.fundmatrixsolution);
    x1i = inlyingMatches{1};
    x2i = inlyingMatches{2};

    if params.verbose
        imc = [img1 img2];
        figure(1); clf; imshow(imc); hold on;
        plot(x1i(1,:),x1i(2,:),'r+');
        plot(x2i(1,:)+size(im1,2),x2i(2,:),'g+');
        for i = 1:length(x1i)
            line([x1i(1,i) x2i(1,i)+size(im1,2)], [x1i(2,i) x2i(2,i)],'color',[0 0 1]);
            %line([m1(2,n) m2(2,n)+size(im1,2)], [m1(1,n) m2(1,n)],'color',[0 0 1])
        end
        hold off
    end
    
    %% calculate 4 relative poses from essential matrix
    E = K2'*F*K1;
    [Pcell] = posesFromEssentialMatrix(E, params.verbose);
    %Ps = getCameraMatrix(E);

    %% resolve ambiguity of the four solutions
    %[P, correctSolution] = cheiralityConstraint(Pcell, K1, K2, x1i, x2i);
    [P, votes] = getCorrectCameraMatrix(cat(3,Pcell{:}), repmat(E,[1,1,4]), K1, K2, [x1i; ones(1,size(x1i,2)); x2i; ones(1,size(x1i,2))]);
    [maxvotes, correctSolution] = max(votes);
    [C R] = getCRfromP(P);

    [rz ry rx] = dcm2angle(R);
    fprintf('Pre-Optimization: pos=(%.4f, %.4f, %.4f), rot=(%.2f, %.2f, %.2f)\n',C,fevalmat(@rad2deg,[rz,ry,rx]));
    
    %% Optimize result
     [C R err] = poseOptimization(x1i, x2i, R, C, K1, K2, 1);
     P = getPfromCR(C,R);
     
    [rz ry rx] = dcm2angle(R);
    fprintf('Post-Optimization: pos=(%.4f, %.4f, %.4f), rot=(%.2f, %.2f, %.2f)\n',C,fevalmat(@rad2deg,[rz,ry,rx]));

    [X dists] = reconstructPoint(x1i,x2i,eye(3),zeros(3,1),R,C,K1,K2);
    err.meanLineDists = mean(dists);
    err.reprojectionErr = reprojectionError(x2i,X,P,K2,2);
    
    if params.verbose
        fh = figure(2);
        clf;
        hold on;
        drawCamera(eye(3),zeros(3,1),0.2,fh,struct('Color','g'));
        drawCamera(R,C,0.2,fh,struct('Color','r'));
        % ---------------------------------------------------------------------
        % Reconstruct points
        % ---------------------------------------------------------------------
        % first: throw out points, that are probably outliers --> simple
        % distance criterion
        Xm = X-repmat(mean(X,2),1,size(X,2));
        % distance to mean:
        Xdists = sqrt(sum(Xm.^2));
        % throw out those, whose distance is greater than 2*std(Xdists)
        X_red = X(:,Xdists < 5*std(Xdists));
        %X_red = X;
        fprintf('Threw out %d points.\n',length(X)-length(X_red));
        scatter3(X_red(1,:),X_red(2,:),X_red(3,:));
        hold off;
        
        %visualizeReprojection(x1i, X, eye(3,4), K1, 2);
        visualizeReprojection(x2i, X, P, K2, 3);
        fprintf('Reprojection error: %.4f\n',err.reprojectionErr);
        
    end

    resultC{iterationCounter} = C;
    resultR{iterationCounter} = R;
    resultErr{iterationCounter} = err;

    
end

fprintf('\n');

if params.repetitions > 1
    BestCamPosition = resultC;
    BestCamRotation = resultR;
    err = resultErr;
else
    BestCamPosition = resultC{1};
    BestCamRotation = resultR{1};
    err = resultErr{1};
end

end



function [params] = parse_parameters(args)
params = struct;
nargs = size(args,2);

% default arguments
params.verbose = false;
params.fundmatrixsolution = '8point';
params.scale = 1;
params.repetitions = 1;
% not enough arguments...
%if (stdargin<2)
%    error('MATLAB:siftfund - Argument parsing: %s', 'You have to specify two images!');
%end
if isempty(args)
    % default parameters
    params.verbose = true;
%     params.img1path = '../data/2010-12-13_Test_JPEG/curtiz_000.jpg';
%     params.img2path = '../data/2010-12-13_Test_JPEG/fassbinder_000.jpg';
%     params.refImgPath = '../data/2010-12-13_Test_JPEG/reference_finca.jpg';
    params.img1path = 'D:\Documents\Uni\SS10\Studienarbeit\repository\codebase/data/canon-home-graffiti/undist_canon_home_01.jpg';
    params.img2path = 'D:\Documents\Uni\SS10\Studienarbeit\repository\codebase/data/canon-home-graffiti/undist_canon_home_08.jpg';
    
    R = angle2dcm(pi,0,0);
    t = [0 150 0]';
    params.extrinsic = [R' -R'*t];
else
    %if ~ischar(varargin{1}) 
    params.img1path = args{1};
    params.img2path = args{2};

    % parse optional arguments
    for i=3:2:nargs
        switch (args{i})
            case 'verbose'
                params.verbose = args{i+1};
            case 'fundmatrixSolution'
                if sum(strcmpi(args{i+1},{'8point','7point','affine'}))>0
                    params.fundmatrixsolution = args{i+1};
                end
            case 'calibrationMatrices'
                params.calibrationMatrices = args{i+1};
            case 'scale'
                params.scale = args{i+1};
            case 'extrinsic'
                params.extrinsic = args{i+1};
            case 'repetitions'
                params.repetitions = args{i+1};
        end
    end
end

end



function [K1 K2 kc1 kc2 img1 img2] = initialize(params)

% resize with "scale"
img1 = imresize(imread(params.img1path),params.scale);
img2 = imresize(imread(params.img2path),params.scale);

if exist('params.calibrationMatrices','var')~=0
    % load calibration matrix
    if iscell(params.calibrationMatrices)
        if length(params.calibrationMatrices)>=2
            K1 = params.calibrationMatrices{1};
            K2 = params.calibrationMatrices{2};
        else
            K1 = params.calibrationMatrices{1};
            K2 = params.calibrationMatrices{1};
        end
    else
        K1 = params.calibrationMatrices;
        K2 = params.calibrationMatrices;
    end
else
    % construct calibration matrix
    %imgdim = size(img1);
    %alpha = deg2rad(48); % ffnungswinkel
    %focal length in px-units
    %f = 0.5/tan(alpha/2)*imgdim(1);
%     K = [ f 0 imgdim(1)/2; ...
%         0 f imgdim(2)/2; ...
%         0 0 1];
    fc = [ 1765.716965252350300 ; 1765.716965252350300 ];
    cc = [ 799.500000000000000 ; 599.500000000000000 ];

    K = [fc(1) 0 cc(1); ...
        0 fc(2) cc(2); ...
        0 0 1];
    
    K1 = K;
    K2 = K;
end

K1 = K1*params.scale; K1(3,3) = 1;
K2 = K2*params.scale; K2(3,3) = 1;

kc1 = zeros(5,1);
kc2 = zeros(5,1);
if exist('params.distortion','var')~=0
    if iscell(params.distortion) && length(params.distortion)==2
        kc1 = params.distortion{1};
        kc2 = params.distortion{2};
    end
end

end

